# -*- coding: utf-8 -*-
import os
import unittest


from ScoutSuite.core.conditions import *

class TestOpinelConditionClass(unittest.TestCase):
    """
    Test opinel.condition
    """

    def test___prepare_age_test(self):
        pass

    def test_pass_condition(self):

        assert pass_condition('a', 'equal', 'a') == True
        assert pass_condition('a', 'equal', 'b') == False
        assert pass_condition(1, 'equal', 1) == True
        assert pass_condition(1, 'equal', 0) == False
        assert pass_condition(('a', 'b'), 'equal', ('a', 'b')) == True
        assert pass_condition(('a', 'b'), 'equal', ('b', 'a')) == False
        assert pass_condition('a', 'notEqual', 'a') == False
        assert pass_condition('a', 'notEqual', 'b') == True
        assert pass_condition(1, 'notEqual', 1) == False
        assert pass_condition(1, 'notEqual', 0) == True
        assert pass_condition(('a', 'b'), 'notEqual', ('a', 'b')) == False
        assert pass_condition(('a', 'b'), 'notEqual', ('b', 'a')) == True

        assert pass_condition(1, 'lessThan', 2) == True
        assert pass_condition(1, 'lessThan', 1) == False
        assert pass_condition(2, 'lessThan', 1) == False
        assert pass_condition(1, 'lessOrEqual', 2) == True
        assert pass_condition(1, 'lessOrEqual', 1) == True
        assert pass_condition(2, 'lessOrEqual', 1) == False
        assert pass_condition(1, 'moreThan', 2) == False
        assert pass_condition(1, 'moreThan', 1) == False
        assert pass_condition(2, 'moreThan', 1) == True
        assert pass_condition(1, 'moreOrEqual', 2) == False
        assert pass_condition(1, 'moreOrEqual', 1) == True
        assert pass_condition(2, 'moreOrEqual', 1) == True

        assert pass_condition({}, 'empty', '') == True
        assert pass_condition({'a': 'b'}, 'empty', '') == False
        assert pass_condition([], 'empty', '') == True
        assert pass_condition([None], 'empty', '') == True
        assert pass_condition(['a'], 'empty', '') == False
        assert pass_condition({}, 'notEmpty', '') == False
        assert pass_condition({'a': 'b'}, 'notEmpty', '') == True
        assert pass_condition([], 'notEmpty', '') == False
        assert pass_condition([None], 'notEmpty', '') == False
        assert pass_condition(['a'], 'notEmpty', '') == True
        assert pass_condition(None, 'null', '') == True
        assert pass_condition('None', 'null', '') == True
        assert pass_condition(None, 'notNull', '') == False
        assert pass_condition('None', 'notNull', '') == False

        assert pass_condition(True, 'true', '') == True
        assert pass_condition('TrUE', 'true', '') == True
        assert pass_condition(False, 'false', '') == True
        assert pass_condition('FaLSe', 'notTrue', '') == True

        test_list1 = []
        test_dict1 = {}
        test_list2 = [
         'a']
        test_dict2 = {'a': 'b'}
        test_list3 = ['a', 'b']
        test_dict3 = {'a': 'b','c': 'd'}
        assert pass_condition(test_list1, 'lengthLessThan', 1) == True
        assert pass_condition(test_list1, 'lengthMoreThan', 1) == False
        assert pass_condition(test_list1, 'lengthEqual', 1) == False
        assert pass_condition(test_list2, 'lengthLessThan', 1) == False
        assert pass_condition(test_list2, 'lengthMoreThan', 1) == False
        assert pass_condition(test_list2, 'lengthEqual', 1) == True
        assert pass_condition(test_list3, 'lengthLessThan', 1) == False
        assert pass_condition(test_list3, 'lengthMoreThan', 1) == True
        assert pass_condition(test_list3, 'lengthEqual', 1) == False
        assert pass_condition(test_dict1, 'lengthLessThan', 1) == True
        assert pass_condition(test_dict1, 'lengthMoreThan', 1) == False
        assert pass_condition(test_dict1, 'lengthEqual', 1) == False
        assert pass_condition(test_dict2, 'lengthLessThan', 1) == False
        assert pass_condition(test_dict2, 'lengthMoreThan', 1) == False
        assert pass_condition(test_dict2, 'lengthEqual', 1) == True
        assert pass_condition(test_dict3, 'lengthLessThan', 1) == False
        assert pass_condition(test_dict3, 'lengthMoreThan', 1) == True
        assert pass_condition(test_dict3, 'lengthEqual', 1) == False

        assert pass_condition(test_dict1, 'withKey', 'a') == False
        assert pass_condition(test_dict2, 'withKey', 'a') == True
        assert pass_condition(test_dict1, 'withoutKey', 'a') == True
        assert pass_condition(test_dict2, 'withoutKey', 'a') == False

        assert pass_condition(test_list1, 'containAtLeastOneOf', test_list1) == False
        assert pass_condition(test_list1, 'containAtLeastOneOf', test_list2) == False
        assert pass_condition(test_list2, 'containAtLeastOneOf', test_list2) == True
        assert pass_condition(test_list2, 'containAtLeastOneOf', ['b']) == False
        assert pass_condition(test_list3, 'containAtLeastOneOf', ['c']) == False
        assert pass_condition(test_list3, 'containAtLeastOneOf', ['c', 'b']) == True
        assert pass_condition('', 'containAtLeastOneOf', test_list1) == False
        assert pass_condition('a', 'containAtLeastOneOf', test_list2) == True
        assert pass_condition(test_list2, 'containAtLeastOneOf', '') == False
        assert pass_condition(test_list2, 'containAtLeastOneOf', 'a') == True
        assert pass_condition(test_list1, 'containAtLeastOneDifferentFrom', test_list1) == False
        assert pass_condition(test_list1, 'containAtLeastOneDifferentFrom', test_list3) == False
        assert pass_condition(test_list2, 'containAtLeastOneDifferentFrom', test_list1) == True
        assert pass_condition(test_list2, 'containAtLeastOneDifferentFrom', test_list2) == False
        assert pass_condition(test_list2, 'containAtLeastOneDifferentFrom', test_list3) == False
        assert pass_condition(['c'], 'containAtLeastOneDifferentFrom', test_list3) == True
        assert pass_condition(test_list3, 'containAtLeastOneDifferentFrom', test_list3) == False
        assert pass_condition(test_list3, 'containAtLeastOneDifferentFrom', test_list2) == True
        assert pass_condition(test_list3, 'containAtLeastOneDifferentFrom', test_list1) == True
        assert pass_condition('', 'containAtLeastOneDifferentFrom', test_list1) == False
        assert pass_condition('a', 'containAtLeastOneDifferentFrom', test_list3) == False
        assert pass_condition('d', 'containAtLeastOneDifferentFrom', test_list3) == True
        assert pass_condition(test_list1, 'containAtLeastOneDifferentFrom', 'a') == False
        assert pass_condition(test_list2, 'containAtLeastOneDifferentFrom', 'd') == True
        assert pass_condition(test_list1, 'containNoneOf', test_list1) == True
        assert pass_condition(test_list2, 'containNoneOf', test_list2) == False
        assert pass_condition(test_list1, 'containNoneOf', 'a') == True
        assert pass_condition('a', 'containNoneOf', test_list1) == True

        assert pass_condition('abcdefg', 'match', '.*cde.*') == True
        assert pass_condition('abcdefg', 'notMatch', '.*cde.*') == False
        assert pass_condition('abcdefg', 'match', '.*345.*') == False
        assert pass_condition('abcdefg', 'notMatch', '.*345.*') == True
        assert pass_condition('abcdefg', 'notMatch', '.*345.*') == True
        assert pass_condition('abcdefg', 'match', ['.*xyx.*', '.*pqr.*']) == False
        assert pass_condition('abcdefg', 'match', ['.*xyx.*', '.*345.*', '.*cde.*']) == True

        assert pass_condition('abcdefg', 'matchInList', '.*cde.*') == True
        assert pass_condition('abcdefg', 'matchInList', '.*edc.*') == False
        assert pass_condition('abcdefg', 'matchInList', ['.*cde.*', '.*edc.*']) == True
        assert pass_condition('abcdefg', 'matchInList', ['.*edc.*', '.*cba.*']) == False
        assert pass_condition(['abcdefg'], 'matchInList', '.*cde.*') == True
        assert pass_condition(['abcdefg'], 'matchInList', '.*edc.*') == False
        assert pass_condition(['abcdefg'], 'matchInList', ['.*cde.*', '.*edc.*']) == True
        assert pass_condition(['abcdefg'], 'matchInList', ['.*edc.*', '.*cba.*']) == False
        assert pass_condition(test_list1, 'matchInList', '.*cde.*') == False
        assert pass_condition('abcdefg', 'matchInList', test_list1) == False
        assert pass_condition(test_list1, 'matchInList', test_list1) == False

        date1 = '2016-04-11 12:20:26.996000+00:00'
        date2 = '2017-04-11 12:20:26.996000+00:00'
        date3 = datetime.datetime.now() - datetime.timedelta(days=1)
        date4 = datetime.datetime.now() - datetime.timedelta(days=100)
        date5 = datetime.datetime.now() - datetime.timedelta(hours=5)
        assert pass_condition(date1, 'priorToDate', date2) == True
        assert pass_condition(date2, 'priorToDate', date1) == False
        assert pass_condition(date3, 'olderThan', [90, 'days']) == False
        assert pass_condition(date4, 'olderThan', [90, 'days']) == True
        assert pass_condition(date5, 'olderThan', [90, 'minutes']) == True
        assert pass_condition(date5, 'olderThan', [360, 'minutes']) == False
        assert pass_condition(date5, 'olderThan', [1, 'hours']) == True
        assert pass_condition(date5, 'olderThan', [6, 'hours']) == False
        try:
            assert pass_condition(date5, 'olderThan', [6, 'unittest']) == False
        except:
            pass

        try:
            assert pass_condition(date5, 'olderThan', '90') == False
        except:
            pass

        assert pass_condition(date3, 'newerThan', [90, 'days']) == True
        assert pass_condition(date4, 'newerThan', [90, 'days']) == False
        assert pass_condition(date4, 'newerThan', ['90', 'days']) == False

        assert pass_condition('192.168.0.1', 'inSubnets', '192.168.0.0/24') == True
        assert pass_condition('192.168.0.1', 'inSubnets', ['192.168.0.0/24']) == True
        assert pass_condition('192.168.1.1', 'inSubnets', ['192.168.0.0/24']) == False
        assert pass_condition('192.168.1.1', 'notInSubnets', ['192.168.0.0/24']) == True

        src_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
        src_file = os.path.join(src_dir, 'policy1.json')

        with open(src_file) as f:
            testpolicy = json.load(f)
        assert pass_condition(testpolicy['Statement'][0], 'containAction', 'iam:GetUser') == True
        assert pass_condition(testpolicy['Statement'][0], 'containAction', 'iam:CreateUser') == False
        assert pass_condition(testpolicy['Statement'][1], 'containAction', 'iam:CreateUser') == True
        assert pass_condition(testpolicy['Statement'][2], 'containAction', 'iam:CreateUser') == True
        assert pass_condition(testpolicy['Statement'][0], 'notContainAction', 'iam:CreateUser') == True
        assert pass_condition(testpolicy['Statement'][0], 'notContainAction', 'iam:GetUser') == False
        assert pass_condition(testpolicy['Statement'][0], 'containAtLeastOneAction', '') == False
        assert pass_condition(testpolicy['Statement'][0], 'containAtLeastOneAction', 'iam:GetUser') == True
        assert pass_condition(testpolicy['Statement'][0], 'containAtLeastOneAction', ['iam:CreateUser', 'iam:GetUser']) == True

        src_file = os.path.join(src_dir, 'statement1.json')
        with open(src_file) as f:
            stringstatement = f.read()
        assert pass_condition(stringstatement, 'containAction', 'iam:GetUser') == True
        assert pass_condition(stringstatement, 'containAtLeastOneAction', 'iam:GetUser') == True

        assert pass_condition('123456789012', 'isSameAccount', '123456789012') == True
        assert pass_condition(['123456789013', '123456789012'], 'isSameAccount', '123456789012') == True
        assert pass_condition('arn:aws:iam::123456789012:root', 'isSameAccount', '123456789012') == True
        assert pass_condition('arn:aws:iam::123456789012:user/name', 'isSameAccount', '123456789012') == True
        assert pass_condition('arn:aws:iam::123456789012:role/name', 'isSameAccount', '123456789012') == True
        assert pass_condition('123456789012', 'isSameAccount', '123456789013') == False
        assert pass_condition('arn:aws:iam::123456789012:root', 'isSameAccount', '123456789013') == False
        assert pass_condition('123456789012', 'isCrossAccount', '123456789013') == True
        assert pass_condition(['123456789013', '123456789012'], 'isCrossAccount', '123456789013') == True
        assert pass_condition('arn:aws:iam::123456789012:root', 'isCrossAccount', '123456789013') == True
        assert pass_condition({'AWS': 'arn:aws:iam::123456789012:root'}, 'isCrossAccount', '123456789013')
        assert pass_condition(
            [{'AWS': 'arn:aws:iam::123456789013:root'}, {'AWS': 'arn:aws:iam::123456789012:root'}],
            'isCrossAccount',
            '123456789013'
        )

        assert pass_condition(["a", "b", "arn:aws:iam::111111111111:role/*"], "containAtLeastOneMatching", ".*[*].*")
        assert pass_condition(["*"], "containAtLeastOneMatching", ".*[*].*")
        assert not pass_condition(["a", "b"], "containAtLeastOneMatching", ".*[*].*")
        assert not pass_condition([], "containAtLeastOneMatching", ".*[*].*")

        try:
            pass_condition('foo', 'bar', 'baz')
        except:
            pass

        return



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.gcp.resources.base import GCPCompositeResources
from ScoutSuite.providers.gcp.resources.cloudsql.backups import Backups
from ScoutSuite.providers.gcp.resources.cloudsql.users import Users
from ScoutSuite.providers.utils import get_non_provider_id


class DatabaseInstances(GCPCompositeResources):
    _children = [
        (Backups, 'backups'),
        (Users, 'users')
    ]

    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_instances = await self.facade.cloudsql.get_database_instances(self.project_id)
        for raw_instance in raw_instances:
            instance_id, instance = self._parse_instance(raw_instance)
            self[instance_id] = instance
        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={instance_id: {'project_id': self.project_id, 'instance_name': instance['name']}
                    for instance_id, instance in self.items()})
        self._set_last_backup_timestamps(self.items())

    def _parse_instance(self, raw_instance):
        instance_dict = {}

        instance_dict['id'] = get_non_provider_id(raw_instance['name'])
        instance_dict['name'] = raw_instance['name']
        instance_dict['project_id'] = raw_instance['project']
        instance_dict['automatic_backup_enabled'] = raw_instance['settings'].get('backupConfiguration', {}).get('enabled')
        instance_dict['database_version'] = raw_instance['databaseVersion']
        instance_dict['log_enabled'] = self._is_log_enabled(raw_instance)
        instance_dict['ssl_required'] = self._is_ssl_required(raw_instance)
        instance_dict['authorized_networks'] = raw_instance['settings'].get('ipConfiguration', {}).get('authorizedNetworks', [])

        if raw_instance['settings'].get('databaseFlags', None):
            instance_dict['local_infile_off'] = self._mysql_local_infile_flag_off(raw_instance)

            instance_dict['log_checkpoints_on'] = self._postgres_flags_on(raw_instance, 'log_checkpoints')
            instance_dict['log_connections_on'] = self._postgres_flags_on(raw_instance, 'log_connections')
            instance_dict['log_disconnections_on'] = self._postgres_flags_on(raw_instance, 'log_disconnections')
            instance_dict['log_lock_waits_on'] = self._postgres_flags_on(raw_instance, 'log_lock_waits')
            instance_dict['log_min_messages'] = self._postgres_log_min_error_statement_flags(raw_instance)
            instance_dict['log_temp_files_0'] = self._postgres_log_temp_files_flags_0(raw_instance)
            instance_dict['log_min_duration_statement_-1'] = self._postgres_log_min_duration_statement_flags_1(
                raw_instance)

            instance_dict['cross_db_ownership_chaining_off'] = self._sqlservers_cross_db_ownership_chaining_flag_off(
                raw_instance, 'cross db ownership chaining')
            instance_dict['contained_database_authentication_off'] = self._sqlservers_cross_db_ownership_chaining_flag_off(
                raw_instance, 'contained database authentication')

        else:
            instance_dict['local_infile_off'] = True

            instance_dict['log_checkpoints_on'] = self._check_database_type(raw_instance)
            instance_dict['log_connections_on'] = self._check_database_type(raw_instance)
            instance_dict['log_disconnections_on'] = self._check_database_type(raw_instance)
            instance_dict['log_lock_waits_on'] = self._check_database_type(raw_instance)
            instance_dict['log_min_messages'] = self._check_database_type(raw_instance)
            instance_dict['log_temp_files_0'] = self._check_database_type(raw_instance)
            instance_dict['log_min_duration_statement_-1'] = self._check_database_type(raw_instance)

            instance_dict['cross_db_ownership_chaining_off'] = True
            instance_dict['contained_database_authentication_off'] = True

        # check if is or has a failover replica
        instance_dict['has_failover_replica'] = raw_instance.get('failoverReplica', []) != []
        instance_dict['is_failover_replica'] = raw_instance.get('masterInstanceName', '') != ''

        # network interfaces
        instance_dict['public_ip'] = None
        instance_dict['private_ip'] = None
        instance_dict['outgoing_ip'] = None
        for address in raw_instance.get('ipAddresses', []):
            if address['type'] == 'PRIMARY':
                instance_dict['public_ip'] = address['ipAddress']
            elif address['type'] == 'PRIVATE':
                instance_dict['private_ip'] = address['ipAddress']
            elif address['type'] == 'OUTGOING':
                instance_dict['outgoing_ip'] = address['ipAddress']
            else:
                print_exception('Unknown Cloud SQL instance IP address type: {}'.format(address['type']))

        return instance_dict['id'], instance_dict

    def _is_log_enabled(self, raw_instance):
        return raw_instance['settings'].get('backupConfiguration', {}).get('binaryLogEnabled')

    def _is_ssl_required(self, raw_instance):
        return raw_instance['settings'].get('ipConfiguration', {}).get('requireSsl', False)

    def _set_last_backup_timestamps(self, instances):
        for instance_id, _ in instances:
            self[instance_id]['last_backup_timestamp'] = self._get_last_backup_timestamp(
                self[instance_id]['backups'])

    def _get_last_backup_timestamp(self, backups):
        if not backups:
            return None
        last_backup_id = max(backups.keys(), key=(
            lambda k: backups[k]['creation_timestamp']))
        return backups[last_backup_id]['creation_timestamp']

    def _mysql_local_infile_flag_off(self, raw_instance):
        if 'MYSQL' in raw_instance['databaseVersion']:
            for flag in raw_instance['settings'].get('databaseFlags', []):
                if flag['name'] == 'local_infile' and flag['value'] == 'on':
                    return False
        return True

    def _check_database_type(self, raw_instance):
        if 'POSTGRES' in raw_instance['databaseVersion']:
            return False
        return None

    def _postgres_flags_on(self, raw_instance, flag_name: str):
        if 'POSTGRES' in raw_instance['databaseVersion']:
            for flag in raw_instance['settings'].get('databaseFlags', []):
                if flag['name'] == flag_name and flag['value'] != 'off':
                    return True
            return False
        else:
            return None

    def _postgres_log_min_error_statement_flags(self, raw_instance):
        if 'POSTGRES' in raw_instance['databaseVersion']:
            for flag in raw_instance['settings'].get('databaseFlags', []):
                if flag['name'] == 'log_min_error_statement' and flag['value'] is not None:
                    return True
            return False
        else:
            return None

    def _postgres_log_temp_files_flags_0(self, raw_instance):
        if 'POSTGRES' in raw_instance['databaseVersion']:
            for flag in raw_instance['settings'].get('databaseFlags', []):
                if flag['name'] == 'log_temp_files' and flag['value'] == 0:
                    return True
            return False
        else:
            return None

    def _postgres_log_min_duration_statement_flags_1(self, raw_instance):
        if 'POSTGRES' in raw_instance['databaseVersion']:
            for flag in raw_instance['settings'].get('databaseFlags', []):
                if flag['name'] == 'log_min_duration_statement' and flag['value'] == -1:
                    return True
            return False
        else:
            return None

    def _sqlservers_cross_db_ownership_chaining_flag_off(self, raw_instance, flag_name: str):
        if 'SQLSERVER' in raw_instance['databaseVersion']:
            for flag in raw_instance['settings'].get('databaseFlags', []):
                if flag['name'] == flag_name and flag['value'] == 'off':
                    return True
            return False
        else:
            return None



from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.core.console import print_exception


class CredentialReports(AWSResources):
    async def fetch_all(self):
        raw_credential_reports = await self.facade.iam.get_credential_reports()
        for raw_credential_report in raw_credential_reports:
            name, resource = await self._parse_credential_reports(raw_credential_report)
            self[name] = resource

    async def _parse_credential_reports(self, raw_credential_report):
        raw_credential_report['id'] = get_non_provider_id(raw_credential_report['user'])
        raw_credential_report['name'] = raw_credential_report['user']
        raw_credential_report['password_enabled'] = raw_credential_report['password_enabled']
        raw_credential_report['password_last_used'] = self._sanitize_date(raw_credential_report['password_last_used'])
        raw_credential_report['password_last_changed'] =\
            self._sanitize_date(raw_credential_report['password_last_changed'])
        raw_credential_report['access_key_1_active'] = raw_credential_report['access_key_1_active']
        raw_credential_report['access_key_1_last_used_date'] =\
            self._sanitize_date(raw_credential_report['access_key_1_last_used_date'])
        raw_credential_report['access_key_1_last_rotated'] = \
            self._sanitize_date(raw_credential_report['access_key_1_last_rotated'])
        raw_credential_report['access_key_2_active'] = raw_credential_report['access_key_2_active']
        raw_credential_report['access_key_2_last_used_date'] =\
            self._sanitize_date(raw_credential_report['access_key_2_last_used_date'])
        raw_credential_report['access_key_2_last_rotated'] = \
            self._sanitize_date(raw_credential_report['access_key_2_last_rotated'])
        raw_credential_report['last_used'] = self._compute_last_used(raw_credential_report)
        raw_credential_report['cert_1_active'] = raw_credential_report['cert_1_active']
        raw_credential_report['cert_2_active'] = raw_credential_report['cert_2_active']

        if raw_credential_report['mfa_active'] == 'true':
            raw_credential_report['mfa_active_hardware'] = await \
                self._user_has_hardware_mfa_devices(raw_credential_report['name'])
        else:
            raw_credential_report['mfa_active_hardware'] = False

        raw_credential_report['partition'] = self.facade.partition

        return raw_credential_report['id'], raw_credential_report

    async def _user_has_hardware_mfa_devices(self, username):
        """
        For a given user, returns whether a hardware MFA device is configured.

        For normal users, virtual devices have serial numbers starting with "arn", so it's easy to validate.

        For the root user, it's not possible to list all the devices, so instead we check all the virtual devices
        to confirm if one is for the root user. If this is not the case, we can infer a hardware device is configured
        (since we know MFA is active for the root user but cannot find a virtual device).
        """
        try:
            if username == '<root_account>':
                devices = await self.facade.iam.get_virtual_mfa_devices()
                for device in devices:
                    # If no EnableDate the device has been disabled
                    if device.get('EnableDate') and device['User']['Arn'][-5:] == ':root':
                        return False
                return True
            else:
                devices = await self.facade.iam.get_user_mfa_devices(username)
                if devices:
                    for device in devices:
                        if device['SerialNumber'][0:4] == 'arn:':
                            return False
                    return True
                else:
                    return False
        except Exception as e:
            print_exception(f'Failed to infer hardware MFA configuration for user {username}: {e}')

    @staticmethod
    def _sanitize_date(date):
        """
        Returns the date if it is not equal to 'N/A' or 'no_information', else returns None
        """
        return date if date != 'no_information' and date != 'N/A' else None

    @staticmethod
    def _compute_last_used(credential_report):
        dates = [credential_report['password_last_used'],
                 credential_report['access_key_1_last_used_date'],
                 credential_report['access_key_2_last_used_date']]

        dates = [date for date in dates if date is not None]
        return max(dates) if len(dates) > 0 else None



from ScoutSuite.providers.azure.facade.loggingmonitoring import LoggingMonitoringFacade
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.kubernetes.authentication_strategy import ClusterProvider, KubernetesCredentials
from ScoutSuite.providers.kubernetes.facade.core import CoreFacade
from ScoutSuite.providers.kubernetes.facade.eks import EKSFacade
from ScoutSuite.providers.kubernetes.facade.extra import ExtraFacade
from ScoutSuite.providers.kubernetes.facade.version import VersionFacade


class KubernetesFacade:
    def __init__(self, credentials: KubernetesCredentials):
        self.eks = None
        self.azure_monitoring = None
        self.gcp = None

        self.core = CoreFacade(credentials)
        self.extra = ExtraFacade(credentials)
        self.version = VersionFacade(credentials)

        if credentials.cluster_provider == ClusterProvider.AKS.value:
            self.azure_monitoring = LoggingMonitoringFacade(credentials)

        elif credentials.cluster_provider == ClusterProvider.EKS.value:
            self.eks = EKSFacade(credentials)

        elif credentials.cluster_provider == ClusterProvider.GKE.value:
            self.gcp = GCPFacade(credentials)



import json
import os
import tempfile
import unittest

from ScoutSuite.core.console import set_logger_configuration, print_error
from ScoutSuite.core.processingengine import ProcessingEngine
from ScoutSuite.core.ruleset import Ruleset


class DummyObject(object):
    pass


class TestScoutRulesProcessingEngine(unittest.TestCase):

    def setUp(self):
        set_logger_configuration(is_debug=True)
        self.rule_counters = {'found': 0, 'tested': 0, 'verified': 0}
        self.test_dir = os.path.dirname(os.path.realpath(__file__))

    # TODO
    # Check that one testcase per finding rule exists (should be within default ruleset)

    def test_all_finding_rules(self):
        ruleset_file_name = os.path.join(self.test_dir, 'data/ruleset-test.json')
        # FIXME this is only for AWS
        with open(os.path.join(self.test_dir, '../ScoutSuite/providers/aws/rules/rulesets/default.json'), 'rt') as f:
            ruleset = json.load(f)

        for rule_file_name in ruleset['rules']:
            self.rule_counters['found'] += 1
            rule = ruleset['rules'][rule_file_name][0]
            rule['enabled'] = True
            print(rule_file_name)
            self._test_rule(ruleset_file_name, rule_file_name, rule)

        print('Existing  rules: %d' % self.rule_counters['found'])
        print('Processed rules: %d' % self.rule_counters['tested'])
        print('Verified  rules: %d' % self.rule_counters['verified'])


    def _test_rule(self, ruleset_file_name, rule_file_name, rule):
        test_config_file_name = os.path.join(self.test_dir, 'data/rule-configs/%s' % rule_file_name)
        if not os.path.isfile(test_config_file_name):
            return
        self.rule_counters['tested'] += 1

        ruleset = self._generate_ruleset(rule_file_name, rule)
        pe = ProcessingEngine(ruleset)

        dummy_provider = DummyObject()
        with open(test_config_file_name, 'rt') as f:
            test_config_dict = json.load(f)
            for key in test_config_dict:
                setattr(dummy_provider, key, test_config_dict[key])
        service = rule_file_name.split('-')[0]
        dummy_provider.service_list = [service]
        pe.run(dummy_provider)
        findings = dummy_provider.services[service]['findings']
        findings = findings[list(findings.keys())[0]]['items']

        test_result_file_name = os.path.join(self.test_dir, 'data/rule-results/%s' % rule_file_name)
        if not os.path.isfile(test_result_file_name):
            print_error('Expected findings:: ')
            print_error(json.dumps(findings, indent=4))
            return

        self.rule_counters['verified'] += 1
        with open(test_result_file_name, 'rt') as f:
            items = json.load(f)

        try:
            assert (set(sorted(findings)) == set(sorted(items)))
        except Exception:
            print_error('Expected items:\n %s' % json.dumps(sorted(items)))
            print_error('Reported items:\n %s' % json.dumps(sorted(findings)))
            assert (False)

    def _generate_ruleset(self, rule_file_name, rule):
        test_ruleset = {'rules': {}, 'about': 'regression test'}
        test_ruleset['rules'][rule_file_name] = [rule]

        with tempfile.NamedTemporaryFile('wt', delete=False) as f:
            f.write(json.dumps(test_ruleset, indent=4))

        return Ruleset(cloud_provider='aws', filename=f.name)



from ScoutSuite.providers.kubernetes.authentication_strategy import KubernetesCredentials
from ScoutSuite.providers.kubernetes.facade.base import KubernetesBaseFacade


class EKSFacade(KubernetesBaseFacade):
    cluster = None

    def __init__(self, credentials: KubernetesCredentials, **kwargs):
        super().__init__(credentials)
        self.context = credentials.cluster_context
        self.session = credentials.aws.session
        self.region = self.session.region_name
        self.eks_client = self.session.client('eks', self.region, **kwargs)

    @KubernetesBaseFacade.continue_upon_exception
    def get_cluster(self, **kwargs):
        if not self.cluster:
            self.cluster = self.eks_client.describe_cluster(name=self.context.split('.')[0], **kwargs)
        return self.cluster['cluster']


from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.utils import get_non_provider_id


class AlertPolicies(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_alert_policies = await self.facade.stackdrivermonitoring.get_alert_policies(self.project_id)
        for raw_alert_policy in raw_alert_policies:
            alert_policy_name, alert_policy = self._parse_alert_policy(raw_alert_policy)
            self[alert_policy_name] = alert_policy

    def _parse_alert_policy(self, raw_alert_policy):
        alert_policy_dict = {}
        alert_policy_dict['id'] = get_non_provider_id(raw_alert_policy.name)
        alert_policy_dict['name'] = raw_alert_policy.display_name
        alert_policy_dict['combiner'] = raw_alert_policy.combiner
        alert_policy_dict['creation_record'] = raw_alert_policy.creation_record
        alert_policy_dict['mutation_record'] = raw_alert_policy.mutation_record
        alert_policy_dict['conditions'] = raw_alert_policy.conditions
        alert_policy_dict['enabled'] = raw_alert_policy.enabled
        return alert_policy_dict['id'], alert_policy_dict





from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.core.console import print_exception


class Buckets(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_buckets = await self.facade.cloudstorage.get_buckets(self.project_id)
        for raw_bucket in raw_buckets:
            bucket_id, bucket = self._parse_bucket(raw_bucket)
            self[bucket_id] = bucket

    def _parse_bucket(self, raw_bucket):
        bucket_dict = {}
        bucket_dict['id'] = get_non_provider_id(raw_bucket.id)
        bucket_dict['name'] = raw_bucket.name
        bucket_dict['project_id'] = self.project_id
        bucket_dict['project_number'] = raw_bucket.project_number
        bucket_dict['creation_date'] = raw_bucket.time_created
        bucket_dict['location'] = raw_bucket.location
        bucket_dict['storage_class'] = raw_bucket.storage_class.lower()
        bucket_dict['versioning_enabled'] = raw_bucket.versioning_enabled
        bucket_dict['logging_enabled'] = raw_bucket.logging is not None

        bucket_dict['public_access_prevention'] = raw_bucket.iam_configuration.public_access_prevention

        iam_configuration = raw_bucket.iam_configuration.get('uniformBucketLevelAccess') or \
            raw_bucket.iam_configuration.get('bucketPolicyOnly')
        if iam_configuration:
            bucket_dict['uniform_bucket_level_access'] = iam_configuration.get("enabled", False)
        else:
            bucket_dict['uniform_bucket_level_access'] = None

        if bucket_dict['uniform_bucket_level_access']:
            bucket_dict['acls'] = []
            bucket_dict['default_object_acl'] = []
        else:
            try:
                bucket_dict['acls'] = list(raw_bucket.acl)
            except Exception as e:
                print_exception(f'Failed to retrieve storage bucket ACLs: {e}')
                bucket_dict['acls'] = []
            try:
                bucket_dict['default_object_acl'] = list(raw_bucket.default_object_acl)
            except Exception as e:
                print_exception(f'Failed to retrieve storage bucket object ACLs: {e}')
                bucket_dict['default_object_acl'] = []

        bucket_dict['member_bindings'] = self._get_cloudstorage_bucket_iam_member_bindings(raw_bucket)

        return bucket_dict['id'], bucket_dict

    def _get_cloudstorage_bucket_iam_member_bindings(self, raw_bucket):
        bucket_iam_policy = raw_bucket.iam_policy
        member_bindings = {}
        if bucket_iam_policy:
            for binding in bucket_iam_policy._bindings:
                for member in binding['members']:
                    if member not in member_bindings:
                        member_bindings[member] = [binding['role']]
                    else:
                        member_bindings[member].append(binding['role'])
        return member_bindings



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class ForwardingRules(Resources):
    def __init__(self, facade: GCPFacade, project_id: str, region: str):
        super().__init__(facade)
        self.project_id = project_id
        self.region = region

    async def fetch_all(self):
        raw_rules = await self.facade.gce.get_forwarding_rules(self.project_id, self.region)
        for raw_rule in raw_rules:
            rule_id, rule = self._parse_forwarding_rule(raw_rule)
            self[rule_id] = rule

    def _parse_forwarding_rule(self, raw_forwarding_rule):
        forwarding_rule_dict = {}
        forwarding_rule_dict['id'] = raw_forwarding_rule.get("id")
        forwarding_rule_dict['name'] = raw_forwarding_rule.get("name")
        forwarding_rule_dict['creation_timestamp'] = raw_forwarding_rule.get("creationTimestamp")
        forwarding_rule_dict['description'] = raw_forwarding_rule.get("description")
        forwarding_rule_dict['region'] = raw_forwarding_rule.get("region")
        forwarding_rule_dict['ip_address'] = raw_forwarding_rule.get("IPAddress")
        forwarding_rule_dict['ip_protocol'] = raw_forwarding_rule.get("IPProtocol")
        forwarding_rule_dict['all_ports'] = raw_forwarding_rule.get("allPorts", False)
        forwarding_rule_dict['port_range'] = raw_forwarding_rule.get("portRange", "")
        forwarding_rule_dict['ports'] = raw_forwarding_rule.get("ports", [])
        forwarding_rule_dict['target'] = raw_forwarding_rule.get("target")
        forwarding_rule_dict['load_balancing_scheme'] = raw_forwarding_rule.get("loadBalancingScheme")
        forwarding_rule_dict['network_tier'] = raw_forwarding_rule.get("networkTier")

        forwarding_rule_dict['subnetwork'] = raw_forwarding_rule.get("subnetwork")
        forwarding_rule_dict['network'] = raw_forwarding_rule.get("network")
        forwarding_rule_dict['backend_service'] = raw_forwarding_rule.get("backendService")
        forwarding_rule_dict['service_label'] = raw_forwarding_rule.get("serviceLabel")
        forwarding_rule_dict['service_name'] = raw_forwarding_rule.get("serviceName")
        forwarding_rule_dict['labels'] = raw_forwarding_rule.get("labels")
        forwarding_rule_dict['ip_version'] = raw_forwarding_rule.get("ipVersion")
        forwarding_rule_dict['allow_global_access'] = raw_forwarding_rule.get("allowGlobalAccess")

        return forwarding_rule_dict['id'], forwarding_rule_dict


from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.providers.azure.utils import get_resource_group_name


class SecurityGroups(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_group in await self.facade.network.get_network_security_groups(self.subscription_id):
            id, network_security_group = self._parse_network_security_group(raw_group)
            self[id] = network_security_group

    def _parse_network_security_group(self, network_security_group):
        network_security_group_dict = {}
        network_security_group_dict['id'] = get_non_provider_id(network_security_group.id)
        network_security_group_dict['name'] = network_security_group.name
        network_security_group_dict['location'] = network_security_group.location
        network_security_group_dict['provisioning_state'] = network_security_group.provisioning_state
        network_security_group_dict['resource_guid'] = network_security_group.resource_guid
        network_security_group_dict['type'] = network_security_group.type
        network_security_group_dict['etag'] = network_security_group.etag
        if network_security_group.tags is not None:
            network_security_group_dict['tags'] = ["{}:{}".format(key, value) for key, value in  network_security_group.tags.items()]
        else:
            network_security_group_dict['tags'] = []
        network_security_group_dict['resource_group_name'] = get_resource_group_name(network_security_group.id)
        network_security_group_dict['additional_properties'] = network_security_group.additional_properties

        network_security_group_dict['security_rules'] = self._parse_security_rules(network_security_group)

        network_security_group_dict['subnets'] = {}
        if network_security_group.subnets:
            for subnet in network_security_group.subnets:
                identifier = get_non_provider_id(subnet.id)
                network_security_group_dict['subnets'][identifier] = {'id': identifier}

        network_security_group_dict['network_interfaces'] = {}
        if network_security_group.network_interfaces:
            for network_interface in network_security_group.network_interfaces:
                identifier = get_non_provider_id(network_interface.id)
                network_security_group_dict['network_interfaces'][identifier] = {'id': identifier}

        return network_security_group_dict['id'], network_security_group_dict

    def _parse_security_rules(self, network_security_group):
        security_rules = {}
        # custom rules
        for sr in network_security_group.security_rules:
            security_rule_id, security_rule_dict = self._parse_security_rule(sr)
            security_rules[security_rule_id] = security_rule_dict
        # default rules
        for sr in network_security_group.default_security_rules:
            security_rule_id, security_rule_dict = self._parse_security_rule(sr, default=True)
            security_rules[security_rule_id] = security_rule_dict
        return security_rules

    def _parse_security_rule(self, rule, default=False):
        security_rule_dict = {}
        security_rule_dict['id'] = rule.id
        security_rule_dict['name'] = rule.name
        security_rule_dict['allow'] = rule.access == "Allow"
        security_rule_dict['priority'] = rule.priority
        security_rule_dict['description'] = rule.description
        security_rule_dict['provisioning_state'] = rule.provisioning_state

        security_rule_dict['protocol'] = rule.protocol
        security_rule_dict['direction'] = rule.direction

        source_address_prefixes = \
            self._merge_prefixes_or_ports(rule.source_address_prefix,
                                          rule.source_address_prefixes if rule.source_address_prefixes else
                                          (get_non_provider_id(rule.source_application_security_groups[0].id) if
                                           rule.source_application_security_groups else None))
        security_rule_dict['source_address_prefixes'] = source_address_prefixes
        # this is required for the HTML partial to interpret the source as an ASG
        if rule.source_application_security_groups:
            security_rule_dict['source_address_prefixes_is_asg'] = True
        else:
            security_rule_dict['source_address_prefixes_is_asg'] = False

        security_rule_dict['source_port_ranges'] = self._merge_prefixes_or_ports(rule.source_port_range, rule.source_port_ranges)
        security_rule_dict['source_ports'] = ['0-65535'] if '*' in security_rule_dict['source_port_ranges'] else security_rule_dict['source_port_ranges']

        security_rule_dict['destination_address_prefixes'] = self._merge_prefixes_or_ports(rule.destination_address_prefix, rule.destination_address_prefixes)

        security_rule_dict['destination_port_ranges'] = self._merge_prefixes_or_ports(rule.destination_port_range, rule.destination_port_ranges)
        security_rule_dict['destination_ports'] = ['0-65535'] if '*' in security_rule_dict['destination_port_ranges'] else security_rule_dict['destination_port_ranges']

        security_rule_dict['etag'] = rule.etag

        security_rule_dict['default'] = default

        return security_rule_dict['id'], security_rule_dict

    def _merge_prefixes_or_ports(self, port_range, port_ranges):
        port_ranges = port_ranges if port_ranges else []
        if port_range:
            port_ranges.append(port_range)
        return port_ranges



from azure.mgmt.keyvault import KeyVaultManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.utils import get_user_agent


class KeyVaultFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = KeyVaultManagementClient(self.credentials.get_credentials(),
                                          subscription_id=subscription_id, user_agent=get_user_agent())
        return client

    async def get_key_vaults(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.vaults.list_by_subscription()))
        except Exception as e:
            print_exception(f'Failed to retrieve key vaults: {e}')
            return []



import asyncio

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import run_concurrently

class CloudFront(AWSBaseFacade):

    async def get_distributions(self):
        client = AWSFacadeUtils.get_client('cloudfront',self.session)
        # When no cloudfront distribution exists, we first need to initiate the creation
        # of a new distributions generate_credential_report by calling
        # client.list_distributions and then check for COMPLETE status before trying to download it:
        aws_cloudfront_api_called, n_attempts = False, 3
        try:
            while not aws_cloudfront_api_called and n_attempts > 0:
                response = await run_concurrently(client.list_distributions)
                if 'ResponseMetadata' in response:
                    aws_cloudfront_api_called = True
                else:
                    n_attempts -= 1
                    await asyncio.sleep(0.1)  # Wait for 100ms before doing a new attempt.
        except Exception as e:
            print_exception('Failed to call aws cloudfront api: {}'.format(e))
            return []
        finally:
            if not aws_cloudfront_api_called and n_attempts == 0:
                print_exception('Failed to call aws cloudfront api in {} attempts'.format(n_attempts))
                return []

        try:
            return response.get('DistributionList', {}).get('Items', [])
        except Exception as e:
            print_exception(f'Failed to get CloudFront distribution lists: {e}')
            return []



import os
import re

from ScoutSuite.core.ruleset import Ruleset
from ScoutSuite.providers.base.provider import BaseProvider
from ScoutSuite.providers.kubernetes.authentication_strategy import ClusterProvider, KubernetesCredentials
from ScoutSuite.providers.kubernetes.resources.workload import Workload
from ScoutSuite.providers.kubernetes.services import KubernetesServicesConfig
from ScoutSuite.utils import formatted_service_name

class KubernetesProvider(BaseProvider):
    """
    Implements provider for Kubernetes
    """

    services_requiring_finding_deduplication = {
        'daemon_set': True,
        'deployment': True,
        'replica_set': True,
        'stateful_set': True,
        'pod': True,
    }

    composite_resources = {
        'loggingmonitoring': True,
        'eks': True,
        'kubernetesengine': True,
        'rbac': True,
        'version': True,
        'workload': True
    }

    def __init__(self, **kwargs):
        report_dir = kwargs.get('report_dir')
        timestamp = kwargs.get('timestamp')
        resources = kwargs.get('resources') or []
        skipped_resources = kwargs.get('skipped_resources') or []
        result_format = kwargs.get('result_format', 'json')

        self.credentials: KubernetesCredentials = kwargs.get('credentials')
        self.metadata_path = f'{os.path.split(os.path.abspath(__file__))[0]}/metadata.json'
        self.environment = 'kubernetes'
        self.provider_code = 'kubernetes'
        self.result_format = result_format
        self.services_config = KubernetesServicesConfig
        self.account_id = self.credentials.cluster_context
        self.provider_name = formatted_service_name.get(self.credentials.cluster_provider) or 'Kubernetes'

        self.original_containers = {
            'cron_job': [],
            'deployment': [],
            'job': [],
            'pod': [],
            'pod_template': [],
            'replica_set': [],
            'stateful_set': [],
        }

        super().__init__(report_dir, timestamp, resources, skipped_resources, result_format)

    def get_report_name(self):
        """
        Returns the name of the report using the provider's configuration
        """
        return f'''kubernetes-{self.credentials.cluster_context.replace(':', '-')}'''

    def preprocessing(self, ip_ranges=None, ip_ranges_name_key=None):
        provider = self.credentials.cluster_provider

        # delete cloud-specific services if necessary
        if provider != ClusterProvider.AKS.value:
            # TODO: have actual AKS findings
            if self.metadata.get(ClusterProvider.AKS.value[0]) and self.metadata[ClusterProvider.AKS.value[0]].get('loggingmonitoring'):
                del self.metadata[ClusterProvider.AKS.value[0]]['loggingmonitoring']
        if provider != ClusterProvider.EKS.value:
            if self.metadata.get(ClusterProvider.EKS.value[0]) and self.metadata[ClusterProvider.EKS.value[0]].get(ClusterProvider.EKS.value):
                del self.metadata[ClusterProvider.EKS.value[0]][ClusterProvider.EKS.value]
        if provider != ClusterProvider.GKE.value:
            if self.metadata.get(ClusterProvider.GKE.value[0]) and self.metadata[ClusterProvider.GKE.value[0]].get('kubernetesengine'):
                del self.metadata[ClusterProvider.GKE.value[0]]['kubernetesengine']

        # delete empty service groups
        service_groups_to_delete = []
        for service_group_name in self.metadata:
            if len(self.metadata[service_group_name]) == 0:
                service_groups_to_delete.append(service_group_name)

        for group_name in service_groups_to_delete:
            del self.metadata[group_name]

        for service_name in Workload.container_path_prefixes:
            keys = Workload.container_path_prefixes[service_name]
            service = self.services.get(service_name)
            if not service: continue
            versions = self._get_resource_versions(service)
            for version in versions:
                resources = service[version]['resources']
                for resource_id in resources:
                    spec = resources[resource_id]
                    for key in keys:
                        spec = spec[key]

                    containers = spec['containers']
                    self.original_containers[service_name] = list(containers)
                    init_containers = spec.get('initContainers', [])
                    ephemeral_containers = spec.get('ephemeralContainers', [])

                    containers.extend(init_containers)
                    containers.extend(ephemeral_containers)

        return super().preprocessing(ip_ranges, ip_ranges_name_key)

    def postprocessing(self, current_time, ruleset: Ruleset, run_parameters):
        self._postprocess_regular_resources()
        self._postprocess_composite_resource('workload')
        self._postprocess_composite_resource('rbac')
        
        for service_name in Workload.container_path_prefixes:
            keys = Workload.container_path_prefixes[service_name]
            service = self.services.get(service_name)
            if not service: continue
            versions = self._get_resource_versions(service)
            for version in versions:
                resources = service[version]['resources']
                for resource_id in resources:
                    spec = resources[resource_id]
                    for key in keys:
                        spec = spec[key]

                    spec['containers'] = self.original_containers[service_name]

        ## TODO: This needs to look better.
        # service_names = self._get_resource_versions(self.services['workload'])
        # for service_name in service_names:
        #     for finding_name in self.services[service_name]['findings']:
        #         self.services['workload']['findings'][finding_name] = self.services[service_name]['findings'][finding_name]

        return super().postprocessing(current_time, ruleset, run_parameters)

    def _get_resource_versions(self, service: dict):
        versions = filter(lambda key: service.get(f'{key}_count') != None, service)
        return list(versions)

    def _load_resource_metadata(self, service_group, service_name, versions):
        self.metadata[service_group] = self.metadata.get(service_group, {})
        self.metadata[service_group][service_name] = self.metadata[service_group].get(service_name, {})
        self.metadata[service_group][service_name]['resources'] = self.metadata[service_group][service_name].get('resources', {})
        self.metadata[service_group][service_name]['summaries'] = self.metadata[service_group][service_name].get('summaries', {})

        for version in versions:
            self.metadata[service_group][service_name]['resources'][version] = {
                'path': f'services.{service_name}.{version}'
            }

    def _postprocess_regular_resources(self):
        for service_name in self.services:
            service_group = service_name[0]
            service = self.services[service_name]

            if self.composite_resources.get(service_name): continue

            service_requires_finding_deduplication = self.services_requiring_finding_deduplication.get(service_name, False)

            versions = self._get_resource_versions(service)
            self._load_resource_metadata(service_group, service_name, versions)

            # post-process findings
            standalone_resources = {}
            standalone_resources_tampered = False

            for version in versions:
                # finding de-duplication
                if not service_requires_finding_deduplication: continue
                service_resources = self.services[service_name][version]['resources']
                for name in service_resources:
                    if not service_resources[name].get('ownerReferences'):
                        standalone_resources[f'''{service_name}.{version}.resources.{name}'''] = True
                        standalone_resources_tampered = True

            # remove resources that have owner references from findings
            findings = self.services[service_name]['findings']
            for finding_name in findings:
                finding = findings[finding_name]

                actual_finding_items = []
                for finding_item in finding['items']:
                    # e.g. pod.v1.resources.pod-name
                    if '.'.join(finding_item.split('.')[:4]) in standalone_resources:
                        actual_finding_items.append(finding_item)

                if standalone_resources_tampered:
                    finding['items'] = actual_finding_items
                    finding['checked_items'] = len(standalone_resources)
                    finding['flagged_items'] = min(finding['checked_items'], len(finding['items']))

                for version in versions:
                    items = finding['items']
                    for i in range(len(items)):
                        expression = f'^{service_name}\.{version}\.resources'
                        items[i] = re.sub(expression, f'{service_name}.{version}', items[i])

    def _postprocess_composite_resource(self, composite_resource_name):
        service_group  = '_scout_suite_aggregation'
        self.metadata[service_group] = self.metadata.get(service_group, {})

        service = self.services[composite_resource_name]
        versions = self._get_resource_versions(service)
        self._load_resource_metadata(service_group, composite_resource_name, versions)


from ScoutSuite.providers.utils import run_concurrently


class GCPFacadeUtils:
    @staticmethod
    async def _get_all(resources, resource_key: str, request, resources_group):
        while request is not None:
            response = request.execute()
            resources.extend(response.get(resource_key, []))
            request = await run_concurrently(
                lambda: resources_group.list_next(previous_request=request, previous_response=response)
            )

    @staticmethod
    async def get_all(resource_key: str, request, resources_group):
        resources = []
        await GCPFacadeUtils._get_all(resources, resource_key, request, resources_group)
        return resources
